@@ -1,8 +1,8 @@
+import sys
 import warnings
 from collections.abc import Mapping, Sequence
 from ipaddress import ip_address
-from urllib.parse import (SplitResult, parse_qsl,
-                          urljoin, urlsplit, urlunsplit)
+from urllib.parse import SplitResult, parse_qsl, urljoin, urlsplit, urlunsplit
 
 from multidict import MultiDict, MultiDictProxy
 import idna
@@ -10,20 +10,12 @@ import idna
 
 from .quoting import _Quoter, _Unquoter
 
-__version__ = '1.1.0'
+__version__ = "1.4.1"
 
-__all__ = ('URL',)
+__all__ = ("URL",)
 
 
-# is_leaf()
-
-
-DEFAULT_PORTS = {
-    'http': 80,
-    'https': 443,
-    'ws': 80,
-    'wss': 443,
-}
+DEFAULT_PORTS = {"http": 80, "https": 443, "ws": 80, "wss": 443}
 
 sentinel = object()
 
@@ -129,92 +121,110 @@ class URL:
     #               / path-noscheme
     #               / path-empty
     # absolute-URI  = scheme ":" hier-part [ "?" query ]
-    __slots__ = ('_cache', '_val')
+    __slots__ = ("_cache", "_val")
 
     _QUOTER = _Quoter()
-    _PATH_QUOTER = _Quoter(safe='@:', protected='/+')
-    _QUERY_QUOTER = _Quoter(safe='?/:@', protected='=+&;', qs=True)
-    _QUERY_PART_QUOTER = _Quoter(safe='?/:@', qs=True)
-    _FRAGMENT_QUOTER = _Quoter(safe='?/:@')
+    _PATH_QUOTER = _Quoter(safe="@:", protected="/+")
+    _QUERY_QUOTER = _Quoter(safe="?/:@", protected="=+&;", qs=True)
+    _QUERY_PART_QUOTER = _Quoter(safe="?/:@", qs=True)
+    _FRAGMENT_QUOTER = _Quoter(safe="?/:@")
 
     _UNQUOTER = _Unquoter()
-    _PATH_UNQUOTER = _Unquoter(unsafe='+')
+    _PATH_UNQUOTER = _Unquoter(unsafe="+")
     _QS_UNQUOTER = _Unquoter(qs=True)
 
-    def __init__(self, val='', *, encoded=False, strict=None):
+    def __new__(cls, val="", *, encoded=False, strict=None):
         if strict is not None:  # pragma: no cover
             warnings.warn("strict parameter is ignored")
-        if isinstance(val, URL):
-            self._val = val._val
-            self._cache = val._cache
-            return
-        if isinstance(val, str):
+        if type(val) is cls:
+            return val
+        if type(val) is str:
             val = urlsplit(val)
-        elif isinstance(val, SplitResult):
+        elif type(val) is SplitResult:
             if not encoded:
                 raise ValueError("Cannot apply decoding to SplitResult")
+        elif isinstance(val, str):
+            val = urlsplit(str(val))
         else:
             raise TypeError("Constructor parameter should be str")
 
         if not encoded:
             if not val[1]:  # netloc
-                netloc = ''
+                netloc = ""
+                host = ""
             else:
                 host = val.hostname
                 if host is None:
-                    raise ValueError(
-                        "Invalid URL: host is required for abolute urls.")
-                netloc = self._make_netloc(val.username,
-                                           val.password,
-                                           host,
-                                           val.port,
-                                           encode=True)
-            path = self._PATH_QUOTER(val[2])
-            if netloc:
-                path = self._normalize_path(path)
+                    raise ValueError("Invalid URL: host is required for absolute urls")
 
-            query = self._QUERY_QUOTER(val[3])
-            fragment = self._FRAGMENT_QUOTER(val[4])
+                try:
+                    port = val.port
+                except ValueError as e:
+                    raise ValueError(
+                        "Invalid URL: port can't be converted to integer"
+                    ) from e
+
+                netloc = cls._make_netloc(
+                    val.username, val.password, host, port, encode=True
+                )
+            path = cls._PATH_QUOTER(val[2])
+            if netloc:
+                path = cls._normalize_path(path)
+
+            cls._validate_authority_uri_abs_path(host=host, path=path)
+            query = cls._QUERY_QUOTER(val[3])
+            fragment = cls._FRAGMENT_QUOTER(val[4])
             val = SplitResult(val[0], netloc, path, query, fragment)
 
+        self = object.__new__(cls)
         self._val = val
         self._cache = {}
+        return self
 
     @classmethod
-    def build(cls, *, scheme='', user='', password='', host='', port=None,
-              path='', query=None, query_string='', fragment='',
-              encoded=False):
+    def build(
+        cls,
+        *,
+        scheme="",
+        user=None,
+        password=None,
+        host="",
+        port=None,
+        path="",
+        query=None,
+        query_string="",
+        fragment="",
+        encoded=False
+    ):
         """Creates and returns a new URL"""
 
-        if host and not scheme:
-            raise ValueError(
-                'Can\'t build URL with "host" but without "scheme".')
         if not host and scheme:
-            raise ValueError(
-                'Can\'t build URL with "scheme" but without "host".')
+            raise ValueError('Can\'t build URL with "scheme" but without "host".')
+        if port and not host:
+            raise ValueError('Can\'t build URL with "port" but without "host".')
         if query and query_string:
-            raise ValueError(
-                "Only one of \"query\" or \"query_string\" should be passed")
+            raise ValueError('Only one of "query" or "query_string" should be passed')
+        if path is None or query_string is None or fragment is None:
+            raise TypeError(
+                'NoneType is illegal for "path", "query_string" and '
+                '"fragment" args, use string values instead.'
+            )
 
-        netloc = cls._make_netloc(user, password, host, port,
-                                  encode=not encoded)
+        if not user and not password and not host and not port:
+            netloc = ""
+        else:
+            netloc = cls._make_netloc(user, password, host, port, encode=not encoded)
         if not encoded:
             path = cls._PATH_QUOTER(path)
             if netloc:
                 path = cls._normalize_path(path)
 
+            cls._validate_authority_uri_abs_path(host=host, path=path)
             query_string = cls._QUERY_QUOTER(query_string)
             fragment = cls._FRAGMENT_QUOTER(fragment)
 
         url = cls(
-            SplitResult(
-                scheme,
-                netloc,
-                path,
-                query_string,
-                fragment
-            ),
-            encoded=True
+            SplitResult(scheme, netloc, path, query_string, fragment), encoded=True
         )
 
         if query:
@@ -222,84 +232,94 @@ class URL:
         else:
             return url
 
+    def __init_subclass__(cls):
+        raise TypeError("Inheritance a class {!r} from URL " "is forbidden".format(cls))
+
     def __str__(self):
         val = self._val
         if not val.path and self.is_absolute() and (val.query or val.fragment):
-            val = val._replace(path='/')
+            val = val._replace(path="/")
         return urlunsplit(val)
 
     def __repr__(self):
         return "{}('{}')".format(self.__class__.__name__, str(self))
 
     def __eq__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
 
         val1 = self._val
         if not val1.path and self.is_absolute():
-            val1 = val1._replace(path='/')
+            val1 = val1._replace(path="/")
 
         val2 = other._val
         if not val2.path and other.is_absolute():
-            val2 = val2._replace(path='/')
+            val2 = val2._replace(path="/")
 
         return val1 == val2
 
     def __hash__(self):
-        ret = self._cache.get('hash')
+        ret = self._cache.get("hash")
         if ret is None:
             val = self._val
             if not val.path and self.is_absolute():
-                val = val._replace(path='/')
-            ret = self._cache['hash'] = hash(val)
+                val = val._replace(path="/")
+            ret = self._cache["hash"] = hash(val)
         return ret
 
     def __le__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val <= other._val
 
     def __lt__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val < other._val
 
     def __ge__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val >= other._val
 
     def __gt__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val > other._val
 
     def __truediv__(self, name):
         name = self._PATH_QUOTER(name)
-        if name.startswith('/'):
-            raise ValueError("Appending path "
-                             "starting from slash is forbidden")
+        if name.startswith("/"):
+            raise ValueError(
+                "Appending path {!r} starting from slash is forbidden".format(name)
+            )
         path = self._val.path
-        if path == '/':
-            new_path = '/' + name
+        if path == "/":
+            new_path = "/" + name
         elif not path and not self.is_absolute():
             new_path = name
         else:
-            parts = path.rstrip('/').split('/')
+            parts = path.rstrip("/").split("/")
             parts.append(name)
-            new_path = '/'.join(parts)
+            new_path = "/".join(parts)
         if self.is_absolute():
             new_path = self._normalize_path(new_path)
-        return URL(self._val._replace(path=new_path, query='', fragment=''),
-                   encoded=True)
+        return URL(
+            self._val._replace(path=new_path, query="", fragment=""), encoded=True
+        )
+
+    def __bool__(self) -> bool:
+        return bool(
+            self._val.netloc or self._val.path or self._val.query or self._val.fragment
+        )
 
     def __getstate__(self):
-        return self._val,
+        return (self._val,)
 
     def __setstate__(self, state):
         if state[0] is None and isinstance(state[1], dict):
             # default style pickle
-            self._val = state[1]['_val']
+            self._val = state[1]["_val"]
         else:
             self._val, *unused = state
         self._cache = {}
@@ -340,9 +360,8 @@ class URL:
         if not self._val.scheme:
             raise ValueError("URL should have scheme")
         v = self._val
-        netloc = self._make_netloc(None, None, v.hostname, v.port,
-                                   encode=False)
-        val = v._replace(netloc=netloc, path='', query='', fragment='')
+        netloc = self._make_netloc(None, None, v.hostname, v.port, encode=False)
+        val = v._replace(netloc=netloc, path="", query="", fragment="")
         return URL(val, encoded=True)
 
     def relative(self):
@@ -353,7 +372,7 @@ class URL:
         """
         if not self.is_absolute():
             raise ValueError("URL should be absolute")
-        val = self._val._replace(scheme='', netloc='')
+        val = self._val._replace(scheme="", netloc="")
         return URL(val, encoded=True)
 
     @property
@@ -426,20 +445,19 @@ class URL:
         raw = self.raw_host
         if raw is None:
             return None
-        if '%' in raw:
+        if "%" in raw:
             # Hack for scoped IPv6 addresses like
             # fe80::2%Проверка
             # presence of '%' sign means only IPv6 address, so idna is useless.
             return raw
-
         try:
-            return idna.decode(raw.encode('ascii'))
+            return idna.decode(raw.encode("ascii"))
         except UnicodeError:  # e.g. '::1'
-            return raw.encode('ascii').decode('idna')
+            return raw.encode("ascii").decode("idna")
 
     @property
     def port(self):
-        """Port part of URL.
+        """Port part of URL, with scheme-based fallback.
 
         None for relative URLs or URLs without explicit port and
         scheme without default port substitution.
@@ -447,6 +465,15 @@ class URL:
         """
         return self._val.port or DEFAULT_PORTS.get(self._val.scheme)
 
+    @property
+    def explicit_port(self):
+        """Port part of URL, without scheme-based fallback.
+
+        None for relative URLs or URLs without explicit port.
+
+        """
+        return self._val.port
+
     @property
     def raw_path(self):
         """Encoded path of URL.
@@ -456,7 +483,7 @@ class URL:
         """
         ret = self._val.path
         if not ret and self.is_absolute():
-            ret = '/'
+            ret = "/"
         return ret
 
     @cached_property
@@ -476,8 +503,7 @@ class URL:
         Empty value if URL has no query part.
 
         """
-        ret = MultiDict(parse_qsl(self.raw_query_string,
-                                  keep_blank_values=True))
+        ret = MultiDict(parse_qsl(self.raw_query_string, keep_blank_values=True))
         return MultiDictProxy(ret)
 
     @property
@@ -503,14 +529,14 @@ class URL:
         """Decoded path of URL with query."""
         if not self.query_string:
             return self.path
-        return '{}?{}'.format(self.path, self.query_string)
+        return "{}?{}".format(self.path, self.query_string)
 
     @cached_property
     def raw_path_qs(self):
         """Encoded path of URL with query."""
         if not self.raw_query_string:
             return self.raw_path
-        return '{}?{}'.format(self.raw_path, self.raw_query_string)
+        return "{}?{}".format(self.raw_path, self.raw_query_string)
 
     @property
     def raw_fragment(self):
@@ -540,14 +566,14 @@ class URL:
         path = self._val.path
         if self.is_absolute():
             if not path:
-                parts = ['/']
+                parts = ["/"]
             else:
-                parts = ['/'] + path[1:].split('/')
+                parts = ["/"] + path[1:].split("/")
         else:
-            if path.startswith('/'):
-                parts = ['/'] + path[1:].split('/')
+            if path.startswith("/"):
+                parts = ["/"] + path[1:].split("/")
             else:
-                parts = path.split('/')
+                parts = path.split("/")
         return tuple(parts)
 
     @cached_property
@@ -566,14 +592,12 @@ class URL:
 
         """
         path = self.raw_path
-        if not path or path == '/':
+        if not path or path == "/":
             if self.raw_fragment or self.raw_query_string:
-                return URL(self._val._replace(query='', fragment=''),
-                           encoded=True)
+                return URL(self._val._replace(query="", fragment=""), encoded=True)
             return self
-        parts = path.split('/')
-        val = self._val._replace(path='/'.join(parts[:-1]),
-                                 query='', fragment='')
+        parts = path.split("/")
+        val = self._val._replace(path="/".join(parts[:-1]), query="", fragment="")
         return URL(val, encoded=True)
 
     @cached_property
@@ -583,7 +607,7 @@ class URL:
         if self.is_absolute():
             parts = parts[1:]
             if not parts:
-                return ''
+                return ""
             else:
                 return parts[-1]
         else:
@@ -594,15 +618,26 @@ class URL:
         """The last part of parts."""
         return self._UNQUOTER(self.raw_name)
 
+    @staticmethod
+    def _validate_authority_uri_abs_path(host, path):
+        """Ensure that path in URL with authority starts with a leading slash.
+
+        Raise ValueError if not.
+        """
+        if len(host) > 0 and len(path) > 0 and not path.startswith("/"):
+            raise ValueError(
+                "Path in a URL with authority " "should start with a slash ('/') if set"
+            )
+
     @classmethod
     def _normalize_path(cls, path):
         # Drop '.' and '..' from path
 
-        segments = path.split('/')
+        segments = path.split("/")
         resolved_path = []
 
         for seg in segments:
-            if seg == '..':
+            if seg == "..":
                 try:
                     resolved_path.pop()
                 except IndexError:
@@ -610,36 +645,62 @@ class URL:
                     # IndexError when popped from resolved_path if
                     # resolving for rfc3986
                     pass
-            elif seg == '.':
+            elif seg == ".":
                 continue
             else:
                 resolved_path.append(seg)
 
-        if segments[-1] in ('.', '..'):
+        if segments[-1] in (".", ".."):
             # do some post-processing here.
             # if the last segment was a relative dir,
             # then we need to append the trailing '/'
-            resolved_path.append('')
+            resolved_path.append("")
 
-        return '/'.join(resolved_path)
+        return "/".join(resolved_path)
 
-    @classmethod
-    def _encode_host(cls, host):
-        try:
-            ip, sep, zone = host.partition('%')
-            ip = ip_address(ip)
-        except ValueError:
+    if sys.version_info >= (3, 7):
+
+        @classmethod
+        def _encode_host(cls, host):
             try:
-                host = idna.encode(host, uts46=True).decode('ascii')
-            except UnicodeError:
-                host = host.encode('idna').decode('ascii')
-        else:
-            host = ip.compressed
-            if sep:
-                host += '%' + zone
-            if ip.version == 6:
-                host = '[' + host + ']'
-        return host
+                ip, sep, zone = host.partition("%")
+                ip = ip_address(ip)
+            except ValueError:
+                # IDNA encoding is slow,
+                # skip it for ASCII-only strings
+                if host.isascii():
+                    return host
+                try:
+                    host = idna.encode(host, uts46=True).decode("ascii")
+                except UnicodeError:
+                    host = host.encode("idna").decode("ascii")
+            else:
+                host = ip.compressed
+                if sep:
+                    host += "%" + zone
+                if ip.version == 6:
+                    host = "[" + host + "]"
+            return host
+
+    else:
+        # the same bug without isascii check
+        @classmethod
+        def _encode_host(cls, host):
+            try:
+                ip, sep, zone = host.partition("%")
+                ip = ip_address(ip)
+            except ValueError:
+                try:
+                    host = idna.encode(host, uts46=True).decode("ascii")
+                except UnicodeError:
+                    host = host.encode("idna").decode("ascii")
+            else:
+                host = ip.compressed
+                if sep:
+                    host += "%" + zone
+                if ip.version == 6:
+                    host = "[" + host + "]"
+            return host
 
     @classmethod
     def _make_netloc(cls, user, password, host, port, encode):
@@ -648,20 +709,20 @@ class URL:
         else:
             ret = host
         if port:
-            ret = ret + ':' + str(port)
-        if password:
+            ret = ret + ":" + str(port)
+        if password is not None:
             if not user:
-                user = ''
+                user = ""
             else:
                 if encode:
                     user = cls._QUOTER(user)
             if encode:
                 password = cls._QUOTER(password)
-            user = user + ':' + password
+            user = user + ":" + password
         elif user and encode:
             user = cls._QUOTER(user)
         if user:
-            ret = user + '@' + ret
+            ret = user + "@" + ret
         return ret
 
     def with_scheme(self, scheme):
@@ -670,8 +731,7 @@ class URL:
         if not isinstance(scheme, str):
             raise TypeError("Invalid scheme type")
         if not self.is_absolute():
-            raise ValueError("scheme replacement is not allowed "
-                             "for relative URLs")
+            raise ValueError("scheme replacement is not allowed " "for relative URLs")
         return URL(self._val._replace(scheme=scheme.lower()), encoded=True)
 
     def with_user(self, user):
@@ -692,14 +752,15 @@ class URL:
         else:
             raise TypeError("Invalid user type")
         if not self.is_absolute():
-            raise ValueError("user replacement is not allowed "
-                             "for relative URLs")
-        return URL(self._val._replace(netloc=self._make_netloc(user,
-                                                               password,
-                                                               val.hostname,
-                                                               val.port,
-                                                               encode=False)),
-                   encoded=True)
+            raise ValueError("user replacement is not allowed " "for relative URLs")
+        return URL(
+            self._val._replace(
+                netloc=self._make_netloc(
+                    user, password, val.hostname, val.port, encode=False
+                )
+            ),
+            encoded=True,
+        )
 
     def with_password(self, password):
         """Return a new URL with password replaced.
@@ -717,17 +778,16 @@ class URL:
         else:
             raise TypeError("Invalid password type")
         if not self.is_absolute():
-            raise ValueError("password replacement is not allowed "
-                             "for relative URLs")
+            raise ValueError("password replacement is not allowed " "for relative URLs")
         val = self._val
         return URL(
             self._val._replace(
-                netloc=self._make_netloc(val.username,
-                                         password,
-                                         val.hostname,
-                                         val.port,
-                                         encode=False)),
-            encoded=True)
+                netloc=self._make_netloc(
+                    val.username, password, val.hostname, val.port, encode=False
+                )
+            ),
+            encoded=True,
+        )
 
     def with_host(self, host):
         """Return a new URL with host replaced.
@@ -742,19 +802,19 @@ class URL:
         if not isinstance(host, str):
             raise TypeError("Invalid host type")
         if not self.is_absolute():
-            raise ValueError("host replacement is not allowed "
-                             "for relative URLs")
+            raise ValueError("host replacement is not allowed " "for relative URLs")
         if not host:
             raise ValueError("host removing is not allowed")
         host = self._encode_host(host)
         val = self._val
         return URL(
-            self._val._replace(netloc=self._make_netloc(val.username,
-                                                        val.password,
-                                                        host,
-                                                        val.port,
-                                                        encode=False)),
-            encoded=True)
+            self._val._replace(
+                netloc=self._make_netloc(
+                    val.username, val.password, host, val.port, encode=False
+                )
+            ),
+            encoded=True,
+        )
 
     def with_port(self, port):
         """Return a new URL with port replaced.
@@ -764,19 +824,18 @@ class URL:
         """
         # N.B. doesn't cleanup query/fragment
         if port is not None and not isinstance(port, int):
-            raise TypeError(
-                "port should be int or None, got {}".format(type(port)))
+            raise TypeError("port should be int or None, got {}".format(type(port)))
         if not self.is_absolute():
-            raise ValueError("port replacement is not allowed "
-                             "for relative URLs")
+            raise ValueError("port replacement is not allowed " "for relative URLs")
         val = self._val
         return URL(
-            self._val._replace(netloc=self._make_netloc(val.username,
-                                                        val.password,
-                                                        val.hostname,
-                                                        port,
-                                                        encode=False)),
-            encoded=True)
+            self._val._replace(
+                netloc=self._make_netloc(
+                    val.username, val.password, val.hostname, port, encode=False
+                )
+            ),
+            encoded=True,
+        )
 
     def with_path(self, path, *, encoded=False):
         """Return a new URL with path replaced."""
@@ -784,51 +843,59 @@ class URL:
             path = self._PATH_QUOTER(path)
             if self.is_absolute():
                 path = self._normalize_path(path)
-        if len(path) > 0 and path[0] != '/':
-            path = '/' + path
-        return URL(self._val._replace(path=path, query='', fragment=''),
-                   encoded=True)
+        if len(path) > 0 and path[0] != "/":
+            path = "/" + path
+        return URL(self._val._replace(path=path, query="", fragment=""), encoded=True)
+
+    @staticmethod
+    def _query_var(v):
+        if isinstance(v, str):
+            return v
+        if type(v) is int:  # no subclasses like bool
+            return str(v)
+        raise TypeError(
+            "Invalid variable type: value "
+            "should be str or int, got {!r} "
+            "of type {}".format(v, type(v))
+        )
 
     def _get_str_query(self, *args, **kwargs):
         if kwargs:
             if len(args) > 0:
-                raise ValueError("Either kwargs or single query parameter "
-                                 "must be present")
+                raise ValueError(
+                    "Either kwargs or single query parameter " "must be present"
+                )
             query = kwargs
         elif len(args) == 1:
             query = args[0]
         else:
-            raise ValueError("Either kwargs or single query parameter "
-                             "must be present")
+            raise ValueError(
+                "Either kwargs or single query parameter " "must be present"
+            )
 
         if query is None:
-            query = ''
+            query = ""
         elif isinstance(query, Mapping):
             quoter = self._QUERY_PART_QUOTER
-            lst = []
-            for k, v in query.items():
-                if isinstance(v, str):
-                    pass
-                elif type(v) == int:  # no subclasses like bool
-                    v = str(v)
-                else:
-                    raise TypeError("Invalid variable type: mapping value "
-                                    "should be str or int, got {!r}".format(v))
-                lst.append(
-                    quoter(k) + '=' + quoter(v))
-            query = '&'.join(lst)
+            query = "&".join(
+                quoter(k) + "=" + quoter(self._query_var(v)) for k, v in query.items()
+            )
         elif isinstance(query, str):
             query = self._QUERY_QUOTER(query)
         elif isinstance(query, (bytes, bytearray, memoryview)):
-            raise TypeError("Invalid query type: bytes, bytearray and "
-                            "memoryview are forbidden")
+            raise TypeError(
+                "Invalid query type: bytes, bytearray and " "memoryview are forbidden"
+            )
         elif isinstance(query, Sequence):
             quoter = self._QUERY_PART_QUOTER
-            query = '&'.join(quoter(k) + '=' + quoter(v)
-                             for k, v in query)
+            query = "&".join(
+                quoter(k) + "=" + quoter(self._query_var(v)) for k, v in query
+            )
         else:
-            raise TypeError("Invalid query type: only str, mapping or "
-                            "sequence of (str, str) pairs is allowed")
+            raise TypeError(
+                "Invalid query type: only str, mapping or "
+                "sequence of (str, str) pairs is allowed"
+            )
 
         return query
 
@@ -838,6 +905,8 @@ class URL:
         Accepts any Mapping (e.g. dict, multidict.MultiDict instances)
         or str, autoencode the argument if needed.
 
+        A sequence of (key, value) pairs is supported as well.
+
         It also can take an arbitrary number of keyword arguments.
 
         Clear query if None is passed.
@@ -847,8 +916,8 @@ class URL:
 
         new_query = self._get_str_query(*args, **kwargs)
         return URL(
-            self._val._replace(path=self._val.path, query=new_query),
-            encoded=True)
+            self._val._replace(path=self._val.path, query=new_query), encoded=True
+        )
 
     def update_query(self, *args, **kwargs):
         """Return a new URL with query part updated."""
@@ -857,8 +926,7 @@ class URL:
         query = MultiDict(self.query)
         query.update(new_query)
 
-        return URL(self._val._replace(query=self._get_str_query(query)),
-                   encoded=True)
+        return URL(self._val._replace(query=self._get_str_query(query)), encoded=True)
 
     def with_fragment(self, fragment):
         """Return a new URL with fragment replaced.
@@ -870,13 +938,14 @@ class URL:
         """
         # N.B. doesn't cleanup query/fragment
         if fragment is None:
-            fragment = ''
+            raw_fragment = ""
         elif not isinstance(fragment, str):
             raise TypeError("Invalid fragment type")
-        return URL(
-            self._val._replace(
-                fragment=self._FRAGMENT_QUOTER(fragment)),
-            encoded=True)
+        else:
+            raw_fragment = self._FRAGMENT_QUOTER(fragment)
+        if self.raw_fragment == raw_fragment:
+            return self
+        return URL(self._val._replace(fragment=raw_fragment), encoded=True)
 
     def with_name(self, name):
         """Return a new URL with name (last part of path) replaced.
@@ -889,10 +958,10 @@ class URL:
         # N.B. DOES cleanup query/fragment
         if not isinstance(name, str):
             raise TypeError("Invalid name type")
-        if '/' in name:
+        if "/" in name:
             raise ValueError("Slash in name is not allowed")
         name = self._PATH_QUOTER(name)
-        if name in ('.', '..'):
+        if name in (".", ".."):
             raise ValueError(". and .. values are forbidden")
         parts = list(self.raw_parts)
         if self.is_absolute():
@@ -900,14 +969,15 @@ class URL:
                 parts.append(name)
             else:
                 parts[-1] = name
-            parts[0] = ''  # replace leading '/'
+            parts[0] = ""  # replace leading '/'
         else:
             parts[-1] = name
-            if parts[0] == '/':
-                parts[0] = ''  # replace leading '/'
+            if parts[0] == "/":
+                parts[0] = ""  # replace leading '/'
         return URL(
-            self._val._replace(path='/'.join(parts),
-                               query='', fragment=''), encoded=True)
+            self._val._replace(path="/".join(parts), query="", fragment=""),
+            encoded=True,
+        )
 
     def join(self, url):
         """Join URLs
@@ -929,12 +999,14 @@ class URL:
     def human_repr(self):
         """Return decoded human readable string for URL representation."""
 
-        return urlunsplit(SplitResult(self.scheme,
-                                      self._make_netloc(self.user,
-                                                        self.password,
-                                                        self.host,
-                                                        self._val.port,
-                                                        encode=False),
-                                      self.path,
-                                      self.query_string,
-                                      self.fragment))
+        return urlunsplit(
+            SplitResult(
+                self.scheme,
+                self._make_netloc(
+                    self.user, self.password, self.host, self._val.port, encode=False
+                ),
+                self.path,
+                self.query_string,
+                self.fragment,
+            )
+        )
